"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const load_features_1 = require("./load-features");
const console_reporter_1 = require("./console-reporter");
const backoff_1 = require("backoff");
const allSuccess = (r, result) => (result.success ? r : false);
exports.afterRx = /^I am run after the "([^"]+)" feature$/;
class FeatureRunner {
    constructor(world, options) {
        this.stepRunners = [];
        this.cleaners = [];
        this.world = world;
        this.featuresDir = options.dir;
        this.reporters = options.reporters || [new console_reporter_1.ConsoleReporter()];
        this.store = options.store || {};
    }
    addStepRunners(runners) {
        this.stepRunners.push(...runners);
        return this;
    }
    cleanup(fn) {
        this.cleaners.push(fn);
    }
    async progress(type, info) {
        await Promise.all(this.reporters.map(reporter => reporter.progress(type, info)));
    }
    async run() {
        const features = await load_features_1.fromDirectory(this.featuresDir);
        const startRun = Date.now();
        const featureResults = [];
        await features.reduce((promise, feature) => promise.then(async () => {
            featureResults.push(await this.runFeature(feature));
        }), Promise.resolve());
        const result = {
            success: featureResults.reduce(allSuccess, true),
            runTime: Date.now() - startRun,
            featureResults,
        };
        await Promise.all(this.reporters.map(reporter => reporter.report(result)));
        await this.cleaners.reduce((promise, cleaner) => promise.then(async () => cleaner(this).then(res => this.progress('cleaner', res))), Promise.resolve());
        return result;
    }
    async runFeature(feature) {
        await this.progress('feature', feature.name);
        if (feature.skip) {
            return {
                feature,
                success: true,
                scenarioResults: [],
            };
        }
        const startRun = Date.now();
        const scenarioResults = [];
        await feature.children.reduce((promise, scenario) => promise.then(async () => {
            if (scenarioResults.length &&
                !scenarioResults[scenarioResults.length - 1].success) {
                scenarioResults.push({
                    success: false,
                    scenario,
                    tries: 0,
                    stepResults: [],
                    skipped: true,
                });
                return;
            }
            if (scenario.type === 'ScenarioOutline' && scenario.examples) {
                const example = scenario.examples[0];
                const header = example.tableHeader.cells.map(({ value }) => value);
                await example.tableBody.reduce((promise, example) => promise.then(async () => {
                    const values = example.cells.map(({ value }) => value);
                    const replace = (str) => header.reduce((str, _, k) => str.replace(`<${header[k]}>`, values[k]), str);
                    const s = {
                        type: 'Scenario',
                        keyword: 'Scenario',
                        argument: scenario.argument,
                        name: `${scenario.name} (${values.join(',')})`,
                        steps: scenario.steps.map(step => (Object.assign({}, step, { text: replace(step.text), argument: step.argument
                                ? Object.assign({}, step.argument, { content: replace(step.argument.content) }) : undefined }))),
                    };
                    scenarioResults.push(await this.retryScenario(s));
                }), Promise.resolve());
            }
            else {
                scenarioResults.push(await this.retryScenario(scenario));
            }
        }), Promise.resolve());
        return {
            success: scenarioResults.reduce(allSuccess, true),
            runTime: Date.now() - startRun,
            feature,
            scenarioResults,
        };
    }
    /**
     * Runs a scenario and retries it with a backoff
     */
    async retryScenario(scenario) {
        return new Promise(async (resolve) => {
            // Run the scenario without delay
            let lastResult = await this.runScenario(scenario);
            if (lastResult.success) {
                return resolve(lastResult);
            }
            // Now retry it, up to 31 seconds
            const b = backoff_1.exponential({
                randomisationFactor: 0,
                initialDelay: 1000,
                maxDelay: 16000,
            });
            b.failAfter(5);
            b.on('ready', async (num) => {
                const r = await this.runScenario(scenario);
                lastResult = Object.assign({}, r, { tries: num + 1 });
                if (lastResult.success) {
                    return resolve(lastResult);
                }
                await this.progress('retry', scenario.name);
                // Retry scenario until timeout
                b.backoff();
            });
            b.on('fail', () => {
                resolve(lastResult);
            });
            b.backoff();
        });
    }
    async runScenario(scenario) {
        await this.progress('scenario', scenario.name);
        const startRun = Date.now();
        const stepResults = [];
        let abort = false;
        await scenario.steps.reduce((promise, step) => promise
            .then(async () => {
            if (abort) {
                stepResults.push({
                    skipped: true,
                    success: false,
                    step: Object.assign({}, step, { interpolatedText: step.text }),
                });
            }
            else {
                stepResults.push(await this.runStep(step));
            }
        })
            .catch(async (error) => {
            await this.progress('step error', error);
            stepResults.push({
                success: false,
                step: Object.assign({}, step, { interpolatedText: step.text }),
                error,
                skipped: false,
            });
            // Skip further steps
            abort = true;
        }), Promise.resolve());
        return {
            success: stepResults.reduce(allSuccess, true),
            runTime: Date.now() - startRun,
            scenario,
            stepResults,
            tries: 1,
            skipped: false,
        };
    }
    async runStep(step) {
        await this.progress('step', step.text);
        const interpolatedStep = Object.assign({}, step, { interpolatedText: this.replaceStoragePlaceholders(step.text), interpolatedArgument: step.argument
                ? this.replaceStoragePlaceholders(step.argument.content)
                : undefined });
        if (exports.afterRx.test(step.text)) {
            return {
                success: true,
                step: interpolatedStep,
                skipped: false,
            };
        }
        const matchedRunner = this.stepRunners.reduce((matchedRunner, runner) => {
            if (matchedRunner) {
                return matchedRunner;
            }
            const args = runner.willRun(interpolatedStep);
            if (args) {
                return {
                    runner,
                    args,
                };
            }
        }, undefined);
        if (!matchedRunner) {
            throw new StepRunnerNotDefinedError(interpolatedStep);
        }
        const startRun = Date.now();
        const result = await matchedRunner.runner.run(matchedRunner.args, interpolatedStep, this);
        return {
            success: true,
            runTime: Date.now() - startRun,
            step: interpolatedStep,
            result,
            skipped: false,
        };
    }
    /**
     * Replace {foo} storage placeholders
     */
    replaceStoragePlaceholders(text) {
        const data = Object.assign({}, this.world, this.store);
        const interpolated = Object.keys(data).reduce((str, key) => str.replace(new RegExp(`{${key}}`, 'g'), data[key]), text);
        const missed = interpolated.match(/\{[^}\W]+\}/g);
        if (missed && missed.length) {
            throw new StoreKeyUndefinedError(missed.map(k => k.slice(1, -1)), data);
        }
        return interpolated;
    }
}
exports.FeatureRunner = FeatureRunner;
class StepRunnerNotDefinedError extends Error {
    constructor(step) {
        super('No runner defined for this step!');
        this.step = step;
        this.name = StepRunnerNotDefinedError.name;
        Error.captureStackTrace(this, StepRunnerNotDefinedError);
        Object.setPrototypeOf(this, StepRunnerNotDefinedError.prototype);
    }
}
exports.StepRunnerNotDefinedError = StepRunnerNotDefinedError;
class RetryError extends Error {
    constructor(step) {
        super('Retrying step failed!');
        this.step = step;
        this.name = RetryError.name;
        Error.captureStackTrace(this, RetryError);
        Object.setPrototypeOf(this, RetryError.prototype);
    }
}
exports.RetryError = RetryError;
class StoreKeyUndefinedError extends Error {
    constructor(keys, store) {
        super(`"${keys.join('"')}" is not defined in the store!`);
        this.keys = keys;
        this.store = store;
        this.name = StoreKeyUndefinedError.name;
        Error.captureStackTrace(this, StoreKeyUndefinedError);
        Object.setPrototypeOf(this, StoreKeyUndefinedError.prototype);
    }
}
exports.StoreKeyUndefinedError = StoreKeyUndefinedError;
