"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const aws_sdk_1 = require("aws-sdk");
const sqs = new aws_sdk_1.SQS();
/**
 * Manages the waiting for webhook requests.
 */
class WebhookReceiver {
    constructor(queueUrl) {
        this.queueUrl = queueUrl;
    }
    /**
     * When multiple alerts are configured, messages may result in the webhook
     * receiver receiving multiple requests from different alerts.
     *
     * This receiver will fetch a webhook request for a specific message group
     * id (which is the path after the API URL: {webhookReceiver}/message-group).
     *
     * Requests from other message groups will be discarded.
     */
    async receiveWebhookRequest(MessageGroupId, runner) {
        const { Messages } = await sqs
            .receiveMessage({
            QueueUrl: this.queueUrl,
            MaxNumberOfMessages: 1,
            MessageAttributeNames: ['All'],
            AttributeNames: ['MessageGroupId'],
            WaitTimeSeconds: 0,
        })
            .promise();
        if (Messages === undefined || !Messages.length) {
            throw new Error('No webhook request received!');
        }
        const { Body, MessageAttributes, ReceiptHandle, Attributes } = Messages[0];
        await sqs
            .deleteMessage({
            QueueUrl: this.queueUrl,
            ReceiptHandle: ReceiptHandle,
        })
            .promise();
        const attrs = MessageAttributes;
        const { MessageGroupId: RcvdMessageGroupId, } = Attributes;
        this.latestWebhookRequest = {
            headers: Object.keys(attrs).reduce((hdrs, key) => {
                hdrs[key] = attrs[key].StringValue;
                return hdrs;
            }, {}),
            body: JSON.parse(Body),
        };
        await runner.progress(`Webhook < ${RcvdMessageGroupId}`, JSON.stringify(this.latestWebhookRequest.body));
        if (RcvdMessageGroupId !== MessageGroupId) {
            throw new Error('Wrong webhook request received!');
        }
        return this.latestWebhookRequest;
    }
    /**
     * Deletes all messages in a Queue instead of using purge (which can only be used every 60 seconds)
     */
    async clearQueue() {
        const { Messages } = await sqs
            .receiveMessage({
            QueueUrl: this.queueUrl,
            MaxNumberOfMessages: 10,
            WaitTimeSeconds: 0,
        })
            .promise();
        if (Messages !== undefined) {
            await Promise.all(Messages.map(({ ReceiptHandle }) => sqs
                .deleteMessage({
                QueueUrl: this.queueUrl,
                ReceiptHandle: ReceiptHandle,
            })
                .promise()));
            await this.clearQueue();
            this.latestWebhookRequest = undefined;
        }
    }
}
exports.WebhookReceiver = WebhookReceiver;
